@@ -281,7 +281,7 @@ int main(int argc, char **argv)
 	/*@only@ */ pvstate_t state = NULL;
 	int retcode = 0;
 	bool can_have_eta = true;
-	bool can_display_utf8 = false;
+	bool terminal_supports_utf8 = false;
 
 #if ! HAVE_SETPROCTITLE
 	initproctitle(argc, argv);
@@ -295,7 +295,7 @@ int main(int argc, char **argv)
 #ifdef HAVE_LANGINFO_H
 	/*@-mustfreefresh@ *//* splint thinks nl_langinfo() leaks memory */
 	if (0 == strcmp(nl_langinfo(CODESET), "UTF-8"))
-		can_display_utf8 = true;
+		terminal_supports_utf8 = true;
 	/*@+mustfreefresh@ */
 #endif
 #endif
@@ -308,16 +308,22 @@ int main(int argc, char **argv)
 	}
 
 	/* Early exit if necessary, such as with "-h". */
-	if (opts->do_nothing) {
+	if (PV_ACTION_NOTHING == opts->action) {
 		debug("%s", "nothing to do - exiting with status 0");
 		opts_free(opts);
 		return 0;
 	}
 
+	/* Set the error message prefix. */
+	/*@-keeptrans@ */
+	pv_set_error_prefix(opts->program_name);
+	/* splint - this function doesn't add an alias or release it. */
+	/*@+keeptrans@ */
+
 	/*
 	 * Allocate our internal state buffer.
 	 */
-	state = pv_state_alloc(opts->program_name);
+	state = pv_state_alloc();
 	if (NULL == state) {
 		/*@-mustfreefresh@ */
 		/*
@@ -381,6 +387,13 @@ int main(int argc, char **argv)
 		pv_state_inputfiles(state, opts->argc, (const char **) (opts->argv));
 	}
 
+	/*
+	 * Put the list of watchfd items into the PV internal state.
+	 */
+	if ((opts->watchfd_count > 0) && (NULL != opts->watchfd_pid) && (NULL != opts->watchfd_fd)) {
+		pv_state_watchfds(state, opts->watchfd_count, opts->watchfd_pid, opts->watchfd_fd);
+	}
+
 	/*
 	 * If stderr is not a terminal and we're neither forcing output nor
 	 * outputting numerically, we will have nothing to display at all.
@@ -452,7 +465,7 @@ int main(int argc, char **argv)
 	pv_state_stop_at_size_set(state, opts->stop_at_size);
 
 	/* Total size calculation, in normal transfer mode. */
-	if (0 == opts->watch_pid) {
+	if (PV_ACTION_TRANSFER == opts->action) {
 		/*
 		 * If no size was given, try to calculate the total size.
 		 */
@@ -505,37 +518,38 @@ int main(int argc, char **argv)
 	pv_state_default_bar_style_set(state, opts->default_bar_style);
 	pv_state_format_string_set(state, opts->format);
 	pv_state_extra_display_set(state, opts->extra_display);
-	pv_state_watch_pid_set(state, opts->watch_pid);
-	pv_state_watch_fd_set(state, opts->watch_fd);
 	pv_state_average_rate_window_set(state, opts->average_rate_window);
 
 	pv_state_set_format(state, opts->progress, opts->timer, can_have_eta ? opts->eta : false,
 			    can_have_eta ? opts->fineta : false, opts->rate, opts->average_rate,
 			    opts->bytes, opts->bufpercent, opts->lastwritten, opts->name);
 
-	debug("%s: %s", "can_display_utf8", can_display_utf8 ? "true" : "false");
-	pv_state_set_can_display_utf8(state, can_display_utf8);
+	debug("%s: %s", "terminal_supports_utf8", terminal_supports_utf8 ? "true" : "false");
+	pv_state_set_terminal_supports_utf8(state, terminal_supports_utf8);
 
 	/* Initialise the signal handling. */
 	pv_sig_init(state);
 
 	/* Run the appropriate main loop. */
-	if (0 == opts->watch_pid && NULL == opts->store_and_forward_file) {
+	switch (opts->action) {
+	case PV_ACTION_NOTHING:
+		break;
+	case PV_ACTION_TRANSFER:
 		/* Normal "transfer data" mode. */
 		pv_remote_init();
 		retcode = pv_main_loop(state);
 		pv_remote_fini();
-	} else if (0 == opts->watch_pid && NULL != opts->store_and_forward_file) {
+		break;
+	case PV_ACTION_STORE_AND_FORWARD:
 		/* Store-and-forward transfer mode. */
 		pv_remote_init();
 		retcode = pv__store_and_forward(state, opts, can_have_eta);
 		pv_remote_fini();
-	} else if (0 != opts->watch_pid && -1 == opts->watch_fd) {
-		/* "Watch all file descriptors of another process" mode. */
-		retcode = pv_watchpid_loop(state);
-	} else if (0 != opts->watch_pid && -1 != opts->watch_fd) {
-		/* "Watch a specific file descriptor of another process" mode. */
+		break;
+	case PV_ACTION_WATCHFD:
+		/* "Watch file descriptor(s) of another process" mode. */
 		retcode = pv_watchfd_loop(state);
+		break;
 	}
 
 	/* Clear up the PID file, if one was written. */
