@@ -1,7 +1,7 @@
 /*
  * Remote-control functions.
  *
- * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2024 Andrew Wood
+ * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2025 Andrew Wood
  *
  * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
@@ -22,7 +22,7 @@
 #include <sys/stat.h>
 
 #ifdef PV_REMOTE_CONTROL
-void pv_error(pvstate_t, char *, ...);
+void pv_error(char *, ...);
 
 struct remote_msg {
 	bool progress;			 /* progress bar flag */
@@ -161,7 +161,7 @@ int pv_remote_set(opts_t opts, pvstate_t state)
 	 * Check that the remote process exists.
 	 */
 	if (kill((pid_t) (opts->remote), 0) != 0) {
-		pv_error(state, "%u: %s", opts->remote, strerror(errno));
+		pv_error("%u: %s", opts->remote, strerror(errno));
 		return PV_ERROREXIT_REMOTE_OR_PID;
 	}
 
@@ -225,7 +225,7 @@ int pv_remote_set(opts_t opts, pvstate_t state)
 	memset(control_filename, 0, sizeof(control_filename));
 	control_fptr = pv__control_file(control_filename, sizeof(control_filename), (pid_t) getpid(), true);
 	if (NULL == control_fptr) {
-		pv_error(state, "%s", strerror(errno));
+		pv_error("%s", strerror(errno));
 		return PV_ERROREXIT_REMOTE_OR_PID;
 	}
 
@@ -234,14 +234,14 @@ int pv_remote_set(opts_t opts, pvstate_t state)
 	 * it.
 	 */
 	if (1 != fwrite(&msgbuf, sizeof(msgbuf), 1, control_fptr)) {
-		pv_error(state, "%s", strerror(errno));
+		pv_error("%s", strerror(errno));
 		(void) fclose(control_fptr);
 		(void) remove(control_filename);
 		return PV_ERROREXIT_REMOTE_OR_PID;
 	}
 
 	if (0 != fclose(control_fptr)) {
-		pv_error(state, "%s", strerror(errno));
+		pv_error("%s", strerror(errno));
 		(void) remove(control_filename);
 		return PV_ERROREXIT_REMOTE_OR_PID;
 	}
@@ -253,7 +253,7 @@ int pv_remote_set(opts_t opts, pvstate_t state)
 	signal_sender = 0;
 	(void) pv_sigusr2_received(state, &signal_sender);
 	if (kill((pid_t) (opts->remote), SIGUSR2) != 0) {
-		pv_error(state, "%u: %s", opts->remote, strerror(errno));
+		pv_error("%u: %s", opts->remote, strerror(errno));
 		(void) remove(control_filename);
 		return PV_ERROREXIT_REMOTE_OR_PID;
 	}
@@ -291,7 +291,7 @@ int pv_remote_set(opts_t opts, pvstate_t state)
 	 * Remove the remote control file.
 	 */
 	if (0 != remove(control_filename)) {
-		pv_error(state, "%s", strerror(errno));
+		pv_error("%s", strerror(errno));
 	}
 
 	/*
@@ -306,7 +306,7 @@ int pv_remote_set(opts_t opts, pvstate_t state)
 	 * warnings, but in this case it's unavoidable, and mitigated by the
 	 * fact we only translate each string once.
 	 */
-	pv_error(state, "%u: %s", opts->remote, _("message not received"));
+	pv_error("%u: %s", opts->remote, _("message not received"));
 	return PV_ERROREXIT_REMOTE_OR_PID;
 	/*@+mustfreefresh @ */
 }
@@ -318,8 +318,10 @@ int pv_remote_set(opts_t opts, pvstate_t state)
  *
  * NB relies on pv_state_set_format() causing the output format to be
  * reparsed.
+ *
+ * Returns true if something was received, false otherwise.
  */
-void pv_remote_check(pvstate_t state)
+bool pv_remote_check(pvstate_t state)
 {
 	pid_t signal_sender;
 	char control_filename[4096];	 /* flawfinder: ignore */
@@ -333,13 +335,13 @@ void pv_remote_check(pvstate_t state)
 	 */
 	signal_sender = 0;
 	if (!pv_sigusr2_received(state, &signal_sender))
-		return;
+		return false;
 
 	memset(control_filename, 0, sizeof(control_filename));
 	control_fptr = pv__control_file(control_filename, sizeof(control_filename), signal_sender, false);
 	if (NULL == control_fptr) {
-		pv_error(state, "%s: %s", control_filename, strerror(errno));
-		return;
+		pv_error("%s: %s", control_filename, strerror(errno));
+		return false;
 	}
 
 	/*
@@ -347,14 +349,14 @@ void pv_remote_check(pvstate_t state)
 	 * it.
 	 */
 	if (1 != fread(&msgbuf, sizeof(msgbuf), 1, control_fptr)) {
-		pv_error(state, "%s", strerror(errno));
+		pv_error("%s", strerror(errno));
 		(void) fclose(control_fptr);
-		return;
+		return false;
 	}
 
 	if (0 != fclose(control_fptr)) {
-		pv_error(state, "%s", strerror(errno));
-		return;
+		pv_error("%s", strerror(errno));
+		return false;
 	}
 
 	/*
@@ -398,6 +400,8 @@ void pv_remote_check(pvstate_t state)
 		pv_state_format_string_set(state, msgbuf.format);
 	if (msgbuf.extra_display[0] != '\0')
 		pv_state_extra_display_set(state, msgbuf.extra_display);
+
+	return true;
 }
 
 
