@@ -21,29 +21,139 @@
 #include <fcntl.h>
 
 
-/* alloc / realloc history buffer */
-static void pv_alloc_history(pvstate_t state)
+/*
+ * Allocate or reallocate the history buffer for calculating average
+ * transfer rate.
+ */
+static void pv_alloc_calc_history(pvtransfercalc_t calc)
 {
-	if (NULL != state->calc.history)
-		free(state->calc.history);
-	state->calc.history = NULL;
+	if (NULL != calc->history)
+		free(calc->history);
+	calc->history = NULL;
 
-	state->calc.history = calloc((size_t) (state->calc.history_len), sizeof(state->calc.history[0]));
-	if (NULL == state->calc.history) {
+	calc->history = calloc((size_t) (calc->history_len), sizeof(calc->history[0]));
+	if (NULL == calc->history) {
 		/*@-mustfreefresh@ */
 		/*
 		 * splint note: the gettext calls made by _() cause memory
 		 * leak warnings, but in this case it's unavoidable, and
 		 * mitigated by the fact we only translate each string once.
 		 */
-		fprintf(stderr, "%s: %s: %s\n", state->status.program_name,
-			_("history structure allocation failed"), strerror(errno));
+		pv_error("%s: %s", _("history structure allocation failed"), strerror(errno));
 		/*@+mustfreefresh@ */
 		return;
 	}
 
-	state->calc.history_first = state->calc.history_last = 0;
-	state->calc.history[0].elapsed_sec = 0.0;	/* to be safe, memset() not recommended for doubles */
+	calc->history_first = calc->history_last = 0;
+	calc->history[0].elapsed_sec = 0.0; /* to be safe, memset() not recommended for doubles */
+}
+
+
+/*
+ * Set the "calc" state sub-structure's history length appropriately for the
+ * given average rate window setting, allocate an appropriate history, and
+ * return the appropriate history interval to use.
+ */
+unsigned int pv_update_calc_average_rate_window(pvtransfercalc_t calc, unsigned int val)
+{
+	unsigned int history_interval;
+
+	if (val < 1)
+		val = 1;
+
+	if (val >= 20) {
+		calc->history_len = (size_t) (val / 5 + 1);
+		history_interval = 5;
+	} else {
+		calc->history_len = (size_t) (val + 1);
+		history_interval = 1;
+	}
+
+	pv_alloc_calc_history(calc);
+
+	return history_interval;
+}
+
+
+/*
+ * Clear the "calc" part of a state structure.
+ */
+void pv_reset_calc(pvtransfercalc_t calc)
+{
+	if (NULL == calc)
+		return;
+	/*
+	 * Explicitly set important floating point values to 0, as memset()
+	 * is not recommended for this.
+	 */
+	calc->transfer_rate = 0.0;
+	calc->average_rate = 0.0;
+	calc->prev_elapsed_sec = 0.0;
+	calc->prev_rate = 0.0;
+	calc->prev_trans = 0.0;
+	calc->current_avg_rate = 0.0;
+	calc->rate_min = 0.0;
+	calc->rate_max = 0.0;
+	calc->rate_sum = 0.0;
+	calc->ratesquared_sum = 0.0;
+	calc->measurements_taken = 0;
+	calc->prev_transferred = 0;
+	calc->percentage = 0.0;
+	calc->history_first = calc->history_last = 0;
+	if (NULL != calc->history) {
+		calc->history[0].elapsed_sec = 0.0;
+	}
+}
+
+
+/*
+ * Clear the "transfer" part of a state structure.
+ */
+void pv_reset_transfer(pvtransferstate_t transfer)
+{
+	if (NULL == transfer)
+		return;
+
+	transfer->elapsed_seconds = 0.0;
+	transfer->read_position = 0;
+	transfer->write_position = 0;
+	transfer->to_write = 0;
+	transfer->written = 0;
+	transfer->total_bytes_read = 0;
+	transfer->total_written = 0;
+	transfer->written_but_not_consumed = 0;
+	transfer->read_errors_in_a_row = 0;
+	transfer->last_read_skip_fd = 0;
+#ifdef HAVE_SPLICE
+	transfer->splice_failed_fd = -1;
+#endif				/* HAVE_SPLICE */
+
+	transfer->line_positions_length = 0;
+	transfer->line_positions_head = 0;
+	transfer->last_output_position = 0;
+}
+
+
+/*
+ * Reset the calculated parts of the "flags" state structure.
+ */
+void pv_reset_flags(pvtransientflags_t flags)
+{
+	if (NULL == flags)
+		return;
+	flags->reparse_display = 1;
+}
+
+
+/*
+ * Reset the calculated parts of a "display" state structure.
+ */
+void pv_reset_display(pvdisplay_t display)
+{
+	if (NULL == display)
+		return;
+	display->initial_offset = 0;
+	display->output_produced = false;
 }
 
 
@@ -55,60 +165,21 @@ void pv_state_reset(pvstate_t state)
 	if (NULL == state)
 		return;
 
-	state->flag.reparse_display = 1;
+	pv_reset_flags(&(state->flags));
 	state->status.current_input_file = -1;
 
-	state->display.initial_offset = 0;
-	state->display.display_visible = false;
+	pv_reset_display(&(state->display));
+	pv_reset_display(&(state->extra_display));
 
-	state->extra_display.initial_offset = 0;
-	state->extra_display.display_visible = false;
-
-	/*
-	 * Explicitly set important floating point values to 0, as memset()
-	 * is not recommended for this.
-	 */
-	state->calc.transfer_rate = 0.0;
-	state->calc.average_rate = 0.0;
-	state->calc.prev_elapsed_sec = 0.0;
-	state->calc.prev_rate = 0.0;
-	state->calc.prev_trans = 0.0;
-	state->calc.current_avg_rate = 0.0;
-	state->calc.rate_min = 0.0;
-	state->calc.rate_max = 0.0;
-	state->calc.rate_sum = 0.0;
-	state->calc.ratesquared_sum = 0.0;
-	state->calc.measurements_taken = 0;
-	state->calc.prev_transferred = 0;
-	state->calc.percentage = 0.0;
-	state->calc.history_first = state->calc.history_last = 0;
-	if (NULL != state->calc.history) {
-		state->calc.history[0].elapsed_sec = 0.0;
-	}
-
-	state->transfer.elapsed_seconds = 0.0;
-	state->transfer.read_position = 0;
-	state->transfer.write_position = 0;
-	state->transfer.to_write = 0;
-	state->transfer.written = 0;
-	state->transfer.total_written = 0;
-	state->transfer.written_but_not_consumed = 0;
-	state->transfer.read_errors_in_a_row = 0;
-	state->transfer.last_read_skip_fd = 0;
-#ifdef HAVE_SPLICE
-	state->transfer.splice_failed_fd = -1;
-#endif				/* HAVE_SPLICE */
-
-	state->transfer.line_positions_length = 0;
-	state->transfer.line_positions_head = 0;
-	state->transfer.last_output_position = 0;
+	pv_reset_calc(&(state->calc));
+	pv_reset_transfer(&(state->transfer));
 }
 
 
 /*
  * Create a new state structure, and return it, or 0 (NULL) on error.
  */
-pvstate_t pv_state_alloc(const char *program_name)
+pvstate_t pv_state_alloc(void)
 {
 	pvstate_t state;
 
@@ -117,19 +188,7 @@ pvstate_t pv_state_alloc(const char *program_name)
 		return NULL;
 	memset(state, 0, sizeof(*state));
 
-	/* splint 3.1.2 thinks this is required for some reason. */
-	if (NULL != state->status.program_name) {
-		free(state->status.program_name);
-	}
-
-	state->status.program_name = pv_strdup(program_name);
-	if (NULL == state->status.program_name) {
-		free(state);
-		return NULL;
-	}
-
-	state->control.watch_pid = 0;
-	state->control.watch_fd = -1;
+	state->watchfd.count = 0;
 	state->control.output_fd = -1;
 #ifdef HAVE_IPC
 	state->cursor.shmid = -1;
@@ -157,6 +216,46 @@ pvstate_t pv_state_alloc(const char *program_name)
 }
 
 
+/*
+ * Free dynamic contents of a display structure.
+ */
+void pv_freecontents_display(pvdisplay_t display)
+{
+	if (NULL != display->display_buffer)
+		free(display->display_buffer);
+	display->display_buffer = NULL;
+}
+
+
+/*
+ * Free dynamic contents of a transfer state structure.
+ */
+void pv_freecontents_transfer(pvtransferstate_t transfer)
+{
+	/*@-keeptrans@ */
+	if (NULL != transfer->transfer_buffer)
+		free(transfer->transfer_buffer);
+	transfer->transfer_buffer = NULL;
+	/*@+keeptrans@ */
+	/* splint - explicitly freeing this structure, so free() here is OK. */
+
+	if (NULL != transfer->line_positions)
+		free(transfer->line_positions);
+	transfer->line_positions = NULL;
+}
+
+
+/*
+ * Free dynamic contents of a calculated transfer state structure.
+ */
+void pv_freecontents_calc(pvtransfercalc_t calc)
+{
+	if (NULL != calc->history)
+		free(calc->history);
+	calc->history = NULL;
+}
+
+
 /*
  * Free a state structure, after which it can no longer be used.
  */
@@ -172,9 +271,9 @@ void pv_state_free(pvstate_t state)
 	if (state->control.output_fd >= 0) {
 		if (STDOUT_FILENO != state->control.output_fd) {
 			if (close(state->control.output_fd) < 0) {
-				fprintf(stderr, "%s: %s: %s\n", state->status.program_name,
-					NULL == state->control.output_name ? "(null)" : state->control.output_name,
-					strerror(errno));
+				pv_error("%s: %s",
+					 NULL == state->control.output_name ? "(null)" : state->control.output_name,
+					 strerror(errno));
 			}
 		}
 		state->control.output_fd = -1;
@@ -185,17 +284,8 @@ void pv_state_free(pvstate_t state)
 		state->control.output_name = NULL;
 	}
 
-	if (NULL != state->status.program_name)
-		free(state->status.program_name);
-	state->status.program_name = NULL;
-
-	if (NULL != state->display.display_buffer)
-		free(state->display.display_buffer);
-	state->display.display_buffer = NULL;
-
-	if (NULL != state->extra_display.display_buffer)
-		free(state->extra_display.display_buffer);
-	state->extra_display.display_buffer = NULL;
+	pv_freecontents_display(&(state->display));
+	pv_freecontents_display(&(state->extra_display));
 
 	if (NULL != state->control.name) {
 		free(state->control.name);
@@ -217,20 +307,9 @@ void pv_state_free(pvstate_t state)
 		state->control.extra_format_string = NULL;
 	}
 
-	/*@-keeptrans@ */
-	if (NULL != state->transfer.transfer_buffer)
-		free(state->transfer.transfer_buffer);
-	state->transfer.transfer_buffer = NULL;
-	/*@+keeptrans@ */
-	/* splint - explicitly freeing this structure, so free() here is OK. */
+	pv_freecontents_transfer(&(state->transfer));
 
-	if (NULL != state->transfer.line_positions)
-		free(state->transfer.line_positions);
-	state->transfer.line_positions = NULL;
-
-	if (NULL != state->calc.history)
-		free(state->calc.history);
-	state->calc.history = NULL;
+	pv_freecontents_calc(&(state->calc));
 
 	if (NULL != state->files.filename) {
 		unsigned int file_idx;
@@ -244,6 +323,15 @@ void pv_state_free(pvstate_t state)
 		state->files.filename = NULL;
 	}
 
+	if (NULL != state->watchfd.pid) {
+		free(state->watchfd.pid);
+		state->watchfd.pid = NULL;
+	}
+	if (NULL != state->watchfd.fd) {
+		free(state->watchfd.fd);
+		state->watchfd.fd = NULL;
+	}
+
 	free(state);
 
 	return;
@@ -316,7 +404,7 @@ void pv_state_set_format(pvstate_t state, bool progress, bool timer, bool eta, b
 		state->control.name = pv_strdup(name);
 
 	/* Tell pv_format() that the format has changed. */
-	state->flag.reparse_display = 1;
+	state->flags.reparse_display = 1;
 }
 
 
@@ -531,16 +619,6 @@ void pv_state_extra_display_set(pvstate_t state, /*@null@ */ const char *val)
 	}
 }
 
-void pv_state_watch_pid_set(pvstate_t state, pid_t val)
-{
-	state->control.watch_pid = val;
-}
-
-void pv_state_watch_fd_set(pvstate_t state, int val)
-{
-	state->control.watch_fd = val;
-}
-
 void pv_state_output_set(pvstate_t state, int fd, const char *name)
 {
 	/*
@@ -549,9 +627,9 @@ void pv_state_output_set(pvstate_t state, int fd, const char *name)
 	 */
 	if (state->control.output_fd >= 0 && state->control.output_fd != STDOUT_FILENO) {
 		if (close(state->control.output_fd) < 0) {
-			fprintf(stderr, "%s: %s: %s\n", state->status.program_name,
-				NULL == state->control.output_name ? "(null)" : state->control.output_name,
-				strerror(errno));
+			pv_error("%s: %s",
+				 NULL == state->control.output_name ? "(null)" : state->control.output_name,
+				 strerror(errno));
 		}
 	}
 	if (NULL != state->control.output_name)
@@ -574,19 +652,12 @@ void pv_state_average_rate_window_set(pvstate_t state, unsigned int val)
 	if (val < 1)
 		val = 1;
 	state->control.average_rate_window = val;
-	if (val >= 20) {
-		state->calc.history_len = (size_t) (val / 5 + 1);
-		state->control.history_interval = 5;
-	} else {
-		state->calc.history_len = (size_t) (val + 1);
-		state->control.history_interval = 1;
-	}
-	pv_alloc_history(state);
+	state->control.history_interval = pv_update_calc_average_rate_window(&(state->calc), val);
 }
 
-void pv_state_set_can_display_utf8(pvstate_t state, bool val)
+void pv_state_set_terminal_supports_utf8(pvstate_t state, bool val)
 {
-	state->control.can_display_utf8 = val;
+	state->status.terminal_supports_utf8 = val;
 }
 
 /*
@@ -595,43 +666,91 @@ void pv_state_set_can_display_utf8(pvstate_t state, bool val)
 void pv_state_inputfiles(pvstate_t state, unsigned int input_file_count, const char **input_files)
 {
 	unsigned int file_idx;
+	/*@only@ */ nullable_string_t *new_array;
 
+	/* Free the old array and its contents, if there was one. */
 	if (NULL != state->files.filename) {
 		for (file_idx = 0; file_idx < state->files.file_count; file_idx++) {
-			/*@-unqualifiedtrans@ */
 			free(state->files.filename[file_idx]);
-			/*@+unqualifiedtrans@ */
-			/*
-			 * TODO: find a way to tell splint the array
-			 * contents are "only" and "null" as well as the
-			 * array itself.
-			 */
 		}
 		free(state->files.filename);
 		state->files.filename = NULL;
 		state->files.file_count = 0;
 	}
-	state->files.filename = calloc((size_t) (input_file_count + 1), sizeof(char *));
-	if (NULL == state->files.filename) {
+
+	/* Allocate an empty new array of the right size. */
+	new_array = calloc((size_t) (input_file_count + 1), sizeof(char *));
+	if (NULL == new_array) {
 		/*@-mustfreefresh@ *//* see similar _() issue above */
-		fprintf(stderr, "%s: %s: %s\n", state->status.program_name, _("file list allocation failed"),
-			strerror(errno));
+		pv_error("%s: %s", _("file list allocation failed"), strerror(errno));
 		/*@+mustfreefresh@ */
 		return;
 	}
+	state->files.filename = new_array;
+
+	/* Populate the new array with copies of the filenames. */
 	for (file_idx = 0; file_idx < input_file_count; file_idx++) {
-		/*@-nullstate@ */
-		state->files.filename[file_idx] = pv_strdup(input_files[file_idx]);
-		if (NULL == state->files.filename[file_idx]) {
+		/*@only@ */ char *new_string;
+		new_string = pv_strdup(input_files[file_idx]);
+		if (NULL == new_string) {
 			/*@-mustfreefresh@ *//* see similar _() issue above */
-			fprintf(stderr, "%s: %s: %s\n", state->status.program_name,
-				_("file list allocation failed"), strerror(errno));
+			pv_error("%s: %s", _("file list allocation failed"), strerror(errno));
 			/*@+mustfreefresh@ */
 			return;
 		}
+		state->files.filename[file_idx] = new_string;
 	}
 	state->files.file_count = input_file_count;
 }
 
-/*@+nullstate@*/
-/* splint: see unqualifiedtrans note by free() above. */
+/*
+ * Set the arrays of watchfd process IDs and file descriptors.
+ */
+void pv_state_watchfds(pvstate_t state, unsigned int watchfd_count, const pid_t * pids, const int *fds)
+{
+	unsigned int item_idx;
+	/*@only@ */ pid_t *new_pid_array = NULL;
+	/*@only@ */ int *new_fd_array = NULL;
+
+	/* Free the old arrays, if there were any. */
+	if (NULL != state->watchfd.pid) {
+		free(state->watchfd.pid);
+		state->watchfd.pid = NULL;
+	}
+	if (NULL != state->watchfd.fd) {
+		free(state->watchfd.fd);
+		state->watchfd.fd = NULL;
+	}
+	state->watchfd.count = 0;
+	state->watchfd.multiple_pids = false;
+
+	/* Allocate empty new arrays of the right size. */
+	new_pid_array = calloc((size_t) (watchfd_count + 1), sizeof(pid_t));
+	if (NULL == new_pid_array) {
+		/*@-mustfreefresh@ *//* see similar _() issue above */
+		pv_error("%s: %s", _("process list allocation failed"), strerror(errno));
+		/*@+mustfreefresh@ */
+		return;
+	}
+	state->watchfd.pid = new_pid_array;
+	new_fd_array = calloc((size_t) (watchfd_count + 1), sizeof(int));
+	if (NULL == new_fd_array) {
+		/*@-mustfreefresh@ *//* see similar _() issue above */
+		pv_error("%s: %s", _("file descriptor list allocation failed"), strerror(errno));
+		/*@+mustfreefresh@ */
+		return;
+	}
+	state->watchfd.fd = new_fd_array;
+
+	/* Populate the new arrays with the values supplied. */
+	for (item_idx = 0; item_idx < watchfd_count; item_idx++) {
+		state->watchfd.pid[item_idx] = pids[item_idx];
+		state->watchfd.fd[item_idx] = fds[item_idx];
+		if ((item_idx > 0) && (pids[item_idx] != pids[item_idx - 1]))
+			state->watchfd.multiple_pids = true;
+	}
+	state->watchfd.count = watchfd_count;
+
+	debug("%s=%d, %s=%s", "watchfd.count", state->watchfd.count, "multiple_pids",
+	      state->watchfd.multiple_pids ? "true" : "false");
+}
