@@ -185,7 +185,7 @@ static ssize_t pv__transfer_read_repeated(int fd, char *buf, size_t count)
  * not guaranteed to succeed for _all_ sizes; we may end up returning 0 if this
  * occurs. (The first write() may return -1 / EINTR if the consumer doesn't
  * read any data before our timeout and the buffer of whatever stdout is is
- * near-full.) (see https://codeberg.org/a-j-wood/pv/pulls/93)
+ * near-full.) (see https://codeberg.org/ivarch/pv/pulls/93)
  *
  * If "sync_after_write" is true, we call fdatasync() after each write() (or
  * fsync() if _POSIX_SYNCHRONIZED_IO is not > 0).
@@ -300,41 +300,63 @@ static ssize_t pv__transfer_write_repeated(int fd, char *buf, size_t count, bool
  * returns 1.
  *
  * At most, the number of bytes read will be the number of bytes remaining
- * in the input buffer.  If state->control.rate_limit is >0, and/or "allowed" is >0,
- * then the maximum number of bytes read will be the number remaining unused
- * in the input buffer or the value of "allowed", whichever is smaller.
+ * in the input buffer, capped to the number of bytes left until
+ * state->control.size is reached if state->control.stop_at_size is true.
+ * If state->control.rate_limit is >0, and/or "max_to_write" is >0, and
+ * splice() is used, then the maximum number of bytes read will be further
+ * capped to the value of "max_to_write", since splice() writes as well as
+ * reads.
  *
- * If splice() was successfully used, sets state->transfer.splice_used to true; if it
- * failed, then state->transfer.splice_failed_fd is updated to the current fd so
- * splice() won't be tried again until the next input file.
+ * If splice() was successfully used, sets state->transfer.splice_used to
+ * true; if it failed, then state->transfer.splice_failed_fd is updated to
+ * the current fd so splice() won't be tried again until the next input
+ * file.
  *
- * Updates state->transfer.read_position by the number of bytes read, unless splice()
- * was used, in which case it does not since there's nothing in the buffer
- * (and it also adds the bytes to state->transfer.written since they've been written
- * to the output).
+ * Updates state->transfer.read_position by the number of bytes read, unless
+ * splice() was used, in which case it does not since there's nothing in the
+ * buffer (and it also adds the bytes to state->transfer.written since
+ * they've been written to the output).
  *
- * On read error, updates state->status.exit_status, and if allowed by
+ * Also increases state->transfer.total_bytes_read by the number of bytes
+ * read, regardless of whether splice() was used, since total_bytes_read is
+ * what it says it is, rather than being a buffer indicator.
+ *
+ * On read error, updates state->status.exit_status, and if max_to_write by
  * state->control.skip_errors, tries to skip past the problem.
  *
  * If the end of the input file is reached or the error is unrecoverable,
  * sets *eof_in to true.  If all data in the buffer has been written at this
  * point, then also sets *eof_out to true.
  */
-static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t allowed)
+static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t max_to_write)
 {
 	bool do_not_skip_errors;
 	size_t bytes_can_read;
 	off_t amount_to_skip, amount_skipped, orig_offset, skip_offset;
 	ssize_t nread;
-#ifdef HAVE_SPLICE
-	size_t bytes_to_splice;
-#endif				/* HAVE_SPLICE */
 
 	do_not_skip_errors = false;
 	if (0 == state->control.skip_errors)
 		do_not_skip_errors = true;
 
 	bytes_can_read = state->transfer.buffer_size - state->transfer.read_position;
+
+	/*
+	 * Don't read past control.size if stop_at_size is true (issue
+	 * #166).
+	 *
+	 * This isn't workable in line mode.
+	 */
+	if (state->control.stop_at_size && !state->control.linemode) {
+		off_t bytes_remaining_to_read = state->control.size - state->transfer.total_bytes_read;
+		if ((long long) bytes_can_read > (long long) bytes_remaining_to_read) {
+			debug("%lld > (%lld-%lld=%lld): %s", (long long) bytes_can_read,
+			      (long long) (state->control.size), (long long) state->transfer.total_bytes_read,
+			      (long long) bytes_remaining_to_read, "truncating for stop-at-size");
+			bytes_can_read = (size_t) bytes_remaining_to_read;
+		}
+	}
+
 	nread = 0;
 
 #ifdef HAVE_SPLICE
@@ -342,8 +364,10 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 	if ((!state->control.linemode) && (!state->control.no_splice)
 	    && (fd != state->transfer.splice_failed_fd)
 	    && (0 == state->transfer.to_write)) {
-		if (state->control.rate_limit > 0 || allowed != 0) {
-			bytes_to_splice = (size_t) allowed;
+		size_t bytes_to_splice;
+
+		if (state->control.rate_limit > 0 || max_to_write != 0) {
+			bytes_to_splice = (size_t) max_to_write;
 		} else {
 			bytes_to_splice = bytes_can_read;
 		}
@@ -365,6 +389,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 			 */
 		} else if (nread > 0) {
 			state->transfer.written = nread;
+			state->transfer.total_bytes_read += nread;
 #ifdef HAVE_FDATASYNC
 			if (state->control.sync_after_write) {
 				/*
@@ -430,6 +455,8 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 #else
 		state->transfer.read_position += nread;
 #endif				/* HAVE_SPLICE */
+		/* Update the counter of all bytes read so far. */
+		state->transfer.total_bytes_read += nread;
 		return 1;
 	}
 
@@ -460,7 +487,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 	 * reached the end of this file.
 	 */
 	if (do_not_skip_errors) {
-		pv_error(state, "%s: %s: %s", pv_current_file_name(state), _("read failed"), strerror(errno));
+		pv_error("%s: %s: %s", pv_current_file_name(state), _("read failed"), strerror(errno));
 		*eof_in = true;
 		if (state->transfer.write_position >= state->transfer.read_position) {
 			*eof_out = true;
@@ -476,7 +503,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 
 	if (!state->transfer.read_error_warning_shown) {
 		/*@-compdef@ */
-		pv_error(state, "%s: %s: %s", pv_current_file_name(state), _("warning: read errors detected"),
+		pv_error("%s: %s: %s", pv_current_file_name(state), _("warning: read errors detected"),
 			 strerror(errno));
 		/*@+compdef@ */
 		/* splint - see previous pv_current_file_name() call. */
@@ -492,7 +519,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 	 */
 	if (0 > orig_offset) {
 		/*@-compdef@ */
-		pv_error(state, "%s: %s: %s", pv_current_file_name(state), _("file is not seekable"), strerror(errno));
+		pv_error("%s: %s: %s", pv_current_file_name(state), _("file is not seekable"), strerror(errno));
 		/*@+compdef@ */
 		/* splint - see previous pv_current_file_name() calls. */
 		*eof_in = true;
@@ -569,8 +596,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 		 */
 		if (EINVAL != errno) {
 			/*@-compdef@ */
-			pv_error(state,
-				 "%s: %s: %s", pv_current_file_name(state), _("failed to seek past error"),
+			pv_error("%s: %s: %s", pv_current_file_name(state), _("failed to seek past error"),
 				 strerror(errno));
 			/*@+compdef@ */
 			/* splint - see previous pv_current_file_name() calls. */
@@ -588,7 +614,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 		state->transfer.read_position += amount_skipped;
 		if (state->control.skip_errors < 2) {
 			/*@-compdef@ */
-			pv_error(state, "%s: %s: %ld - %ld (%ld %s)",
+			pv_error("%s: %s: %ld - %ld (%ld %s)",
 				 pv_current_file_name(state),
 				 _("skipped past read error"), (long) orig_offset, (long) skip_offset,
 				 (long) amount_skipped, _("B"));
@@ -631,7 +657,7 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 	int write_errno;
 
 	if (NULL == state->transfer.transfer_buffer) {
-		pv_error(state, "%s", _("no transfer buffer allocated"));
+		pv_error("%s", _("no transfer buffer allocated"));
 		state->status.exit_status |= PV_ERROREXIT_MEMORY;
 		*eof_out = true;
 		state->transfer.written = -1;
@@ -673,7 +699,7 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 		      (long) (new_timer.it_value.tv_usec));
 
 		if (0 != setitimer(ITIMER_REAL, &new_timer, NULL)) {
-			pv_error(state, "%s: %s", _("failed to set interval timer"), strerror(errno));
+			pv_error("%s: %s", _("failed to set interval timer"), strerror(errno));
 		}
 
 #else				/* ! HAVE_SETITIMER */
@@ -699,7 +725,7 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 		new_timer.it_value.tv_sec = 0;
 		new_timer.it_value.tv_usec = 0;
 		if (0 != setitimer(ITIMER_REAL, &new_timer, NULL)) {
-			pv_error(state, "%s: %s", _("failed to clear interval timer"), strerror(errno));
+			pv_error("%s: %s", _("failed to clear interval timer"), strerror(errno));
 		}
 
 		/*@+unrecog@ */
@@ -741,7 +767,7 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 				state->transfer.line_positions =
 				    calloc((size_t) (state->transfer.line_positions_capacity), sizeof(off_t));
 				if (NULL == state->transfer.line_positions) {
-					pv_error(state, "%s: %s", _("line position buffer allocation failed"),
+					pv_error("%s: %s", _("line position buffer allocation failed"),
 						 strerror(errno));
 				}
 				/*@+mustfreeonly@ */
@@ -910,11 +936,12 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 	if (EPIPE == write_errno) {
 		*eof_in = true;
 		*eof_out = true;
-		state->flag.pipe_closed = 1;
+		state->flags.pipe_closed = 1;
+		debug("%s", "SIGPIPE received - setting pipe_closed");
 		return 0;
 	}
 
-	pv_error(state, "%s: %s", _("write failed"), strerror(write_errno));
+	pv_error("%s: %s", _("write failed"), strerror(write_errno));
 	state->status.exit_status |= PV_ERROREXIT_TRANSFER;
 	*eof_out = true;
 	state->transfer.written = -1;
@@ -939,7 +966,7 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 /*@only@*/
 static char *pv__allocate_aligned_buffer(int outfd, int infd, size_t target_size)
 {
-	char *newptr;
+	void *newptr;
 
 #if defined(HAVE_FPATHCONF) && defined(HAVE_POSIX_MEMALIGN) && defined(_PC_REC_XFER_ALIGN)
 	long input_alignment, output_alignment, min_alignment;
@@ -1053,7 +1080,7 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 		state->transfer.transfer_buffer =
 		    pv__allocate_aligned_buffer(state->control.output_fd, fd, state->control.target_buffer_size + 32);
 		if (NULL == state->transfer.transfer_buffer) {
-			pv_error(state, "%s: %s", _("buffer allocation failed"), strerror(errno));
+			pv_error("%s: %s", _("buffer allocation failed"), strerror(errno));
 			state->status.exit_status |= PV_ERROREXIT_MEMORY;
 			return -1;
 		}
@@ -1101,8 +1128,10 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 	if ((state->control.linemode) && (lineswritten != NULL))
 		*lineswritten = 0;
 
-	if ((*eof_in) && (*eof_out))
+	if ((*eof_in) && (*eof_out)) {
+		debug("%s %d: %s", "fd", fd, "early return 0 - EOF in and out");
 		return 0;
+	}
 
 	check_read_fd = -1;
 	check_write_fd = -1;
@@ -1144,15 +1173,16 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 		/*
 		 * Ignore transient errors by returning 0 immediately.
 		 */
-		if (EINTR == errno)
+		if (EINTR == errno) {
+			debug("%s %d: %s", "fd", fd, "early return 0 - is_data_ready < 0");
 			return 0;
+		}
 
 		/*
 		 * Any other error is a problem and we must report back.
 		 */
 		/*@-compdef@ */
-		pv_error(state, "%s: %s: %d: %s", pv_current_file_name(state), _("select call failed"), n,
-			 strerror(errno));
+		pv_error("%s: %s: %d: %s", pv_current_file_name(state), _("select call failed"), n, strerror(errno));
 		/*@+compdef@ */
 		/* splint - see previous pv_current_file_name() calls. */
 
@@ -1170,8 +1200,12 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 	 * NB this can update state->transfer.written because of splice().
 	 */
 	if (ready_to_read) {
-		if (pv__transfer_read(state, fd, eof_in, eof_out, allowed) == 0)
+		if (pv__transfer_read(state, fd, eof_in, eof_out, allowed) == 0) {
+			debug("%s %d: %s (%s=%s, %s=%s, %s=%lu)", "fd", fd,
+			      "early return 0 - pv__transfer_read returned 0", "eof_in", eof_in ? "true" : "false",
+			      "eof_out", eof_out ? "true" : "false", "allowed", (unsigned long) allowed);
 			return 0;
+		}
 	}
 
 	/*
@@ -1202,8 +1236,12 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 	    && (state->transfer.read_position > state->transfer.write_position)
 	    && (state->transfer.to_write > 0)
 	    && (NULL != lineswritten)) {
-		if (pv__transfer_write(state, eof_in, eof_out, lineswritten) == 0)
+		if (pv__transfer_write(state, eof_in, eof_out, lineswritten) == 0) {
+			debug("%s %d: %s (%s=%s, %s=%s, %s=%lu)", "fd", fd,
+			      "early return 0 - pv__transfer_write returned 0", "eof_in", eof_in ? "true" : "false",
+			      "eof_out", eof_out ? "true" : "false", "lineswritten", (unsigned long) lineswritten);
 			return 0;
+		}
 	}
 #ifdef MAXIMISE_BUFFER_FILL
 	/*
@@ -1225,5 +1263,9 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 	}
 #endif				/* MAXIMISE_BUFFER_FILL */
 
+	if (0 == state->transfer.written) {
+		debug("%s %d: %s", "fd", fd, "end-of-function return 0 - transfer.written is zero");
+	}
+
 	return state->transfer.written;
 }
