@@ -96,7 +96,7 @@ int pv_watchfd_info(pvstate_t state, pvwatchfd_t info, bool automatic)
 
 	if (kill(info->watch_pid, 0) != 0) {
 		if (!automatic)
-			pv_error(state, "%s %u: %s", _("pid"), info->watch_pid, strerror(errno));
+			pv_error("%s %u: %s", _("pid"), info->watch_pid, strerror(errno));
 		return 1;
 	}
 
@@ -105,8 +105,7 @@ int pv_watchfd_info(pvstate_t state, pvwatchfd_t info, bool automatic)
 				  PROC_PIDFDVNODEPATHINFO, &vnodeInfo,
 				  PROC_PIDFDVNODEPATHINFO_SIZE);
 	if (size != PROC_PIDFDVNODEPATHINFO_SIZE) {
-		pv_error(state, "%s %u: %s %d: %s",
-			 _("pid"), info->watch_pid, _("fd"), info->watch_fd, strerror(errno));
+		pv_error("%s %u: %s %d: %s", _("pid"), info->watch_pid, _("fd"), info->watch_fd, strerror(errno));
 		return 3;
 	}
 
@@ -116,7 +115,7 @@ int pv_watchfd_info(pvstate_t state, pvwatchfd_t info, bool automatic)
 
 	if (!(0 == stat(info->file_fdpath, &(info->sb_fd)))) {
 		if (!automatic)
-			pv_error(state, "%s %u: %s %d: %s: %s",
+			pv_error("%s %u: %s %d: %s: %s",
 				 _("pid"),
 				 info->watch_pid, _("fd"), info->watch_fd, info->file_fdpath, strerror(errno));
 		return 3;
@@ -124,7 +123,7 @@ int pv_watchfd_info(pvstate_t state, pvwatchfd_t info, bool automatic)
 
 	if (!filesize(info)) {
 		if (!automatic)
-			pv_error(state, "%s %u: %s %d: %s: %s",
+			pv_error("%s %u: %s %d: %s: %s",
 				 _("pid"),
 				 info->watch_pid,
 				 _("fd"), info->watch_fd, info->file_fdpath, _("not a regular file or block device"));
@@ -162,7 +161,7 @@ int pv_watchfd_info(pvstate_t state, pvwatchfd_t info, bool automatic)
 
 	if (kill(info->watch_pid, 0) != 0) {
 		if (!automatic)
-			pv_error(state, "%s %u: %s", _("pid"), info->watch_pid, strerror(errno));
+			pv_error("%s %u: %s", _("pid"), info->watch_pid, strerror(errno));
 		return 1;
 	}
 	(void) pv_snprintf(info->file_fdinfo, PV_SIZEOF_FILE_FDINFO,
@@ -180,7 +179,7 @@ int pv_watchfd_info(pvstate_t state, pvwatchfd_t info, bool automatic)
 		 * and then only if it's a block device - see filesize().
 		 */
 		if (!automatic)
-			pv_error(state, "%s %u: %s %d: %s",
+			pv_error("%s %u: %s %d: %s",
 				 _("pid"), info->watch_pid, _("fd"), info->watch_fd, strerror(errno));
 		return 2;
 	}
@@ -188,7 +187,7 @@ int pv_watchfd_info(pvstate_t state, pvwatchfd_t info, bool automatic)
 	if (!((0 == stat(info->file_fd, &(info->sb_fd)))
 	      && (0 == lstat(info->file_fd, &(info->sb_fd_link))))) {
 		if (!automatic)
-			pv_error(state, "%s %u: %s %d: %s: %s",
+			pv_error("%s %u: %s %d: %s: %s",
 				 _("pid"),
 				 info->watch_pid, _("fd"), info->watch_fd, info->file_fdpath, strerror(errno));
 		return 3;
@@ -198,7 +197,7 @@ int pv_watchfd_info(pvstate_t state, pvwatchfd_t info, bool automatic)
 
 	if (!filesize(info)) {
 		if (!automatic)
-			pv_error(state, "%s %u: %s %d: %s: %s",
+			pv_error("%s %u: %s %d: %s: %s",
 				 _("pid"),
 				 info->watch_pid,
 				 _("fd"), info->watch_fd, info->file_fdpath, _("not a regular file or block device"));
@@ -297,7 +296,7 @@ static int pidfds(pvstate_t state, unsigned int pid, struct proc_fdinfo **fds, i
 {
 	int size_needed = proc_pidinfo(pid, PROC_PIDLISTFDS, 0, 0, 0);
 	if (size_needed == -1) {
-		pv_error(state, "%s: unable to list pid fds: %s", _("pid"), strerror(errno));
+		pv_error("%s: unable to list pid fds: %s", _("pid"), strerror(errno));
 		return -1;
 	}
 
@@ -305,7 +304,7 @@ static int pidfds(pvstate_t state, unsigned int pid, struct proc_fdinfo **fds, i
 
 	*fds = (struct proc_fdinfo *) malloc(size_needed);
 	if (*fds == NULL) {
-		pv_error(state, "%s: alloc failed: %s", _("pid"), strerror(errno));
+		pv_error("%s: alloc failed: %s", _("pid"), strerror(errno));
 		return -1;
 	}
 	memset(*fds, 0, size_needed);
@@ -354,6 +353,10 @@ static bool extend_info_array(int *array_length_ptr, pvwatchfd_t * info_array_pt
 		return false;
 	}
 
+	new_info_array[array_length - 1].unused = true;
+
+	debug("%s", "extended info array");
+
 	*info_array_ptr = new_info_array;
 	*array_length_ptr = array_length;
 	return true;
@@ -364,25 +367,76 @@ static bool extend_info_array(int *array_length_ptr, pvwatchfd_t * info_array_pt
 /*@+compdef @ */
 
 
+/*
+ * Reset calculated values in the given watchfd info structure.
+ */
+void pv_reset_watchfd(pvwatchfd_t info)
+{
+	if (NULL == info)
+		return;
+	pv_reset_calc(&(info->calc));
+	pv_reset_transfer(&(info->transfer));
+	pv_reset_flags(&(info->flags));
+	pv_reset_display(&(info->display));
+}
+
+
+/*
+ * Free dynamically allocated areas in the given watchfd info structure.
+ */
+void pv_freecontents_watchfd(pvwatchfd_t info)
+{
+	if (NULL == info)
+		return;
+	pv_freecontents_calc(&(info->calc));
+	pv_freecontents_transfer(&(info->transfer));
+	pv_freecontents_display(&(info->display));
+}
+
+
+/*
+ * Comparison function for qsort, to compare two pvwatchfd_s structures.
+ */
+static int pv_compare_watchfd(const void *a, const void *b)
+{
+	int fd_a, fd_b;
+
+	fd_a = 0;
+	fd_b = 0;
+	if (NULL != a)
+		fd_a = ((pvwatchfd_t) a)->watch_fd;
+	if (NULL != b)
+		fd_b = ((pvwatchfd_t) b)->watch_fd;
+
+	if (fd_a < fd_b)
+		return -1;
+	if (fd_a > fd_b)
+		return 1;
+	return 0;
+}
+
+
 /*
  * Scan the given process and update the arrays with any new file
- * descriptors.
+ * descriptors.  If "watch_fd" is not -1, then all other file descriptors
+ * numbers will be ignored,
  *
  * Returns 0 on success, 1 if the process no longer exists or could not be
  * read, or 2 for a memory allocation error.
  */
-int pv_watchpid_scanfds(pvstate_t state,
-			pid_t watch_pid, int *array_length_ptr, pvwatchfd_t * info_array_ptr, int *fd_to_idx)
+int pv_watchpid_scanfds(pvstate_t state, pid_t watch_pid, int watch_fd, int *array_length_ptr,
+			pvwatchfd_t * info_array_ptr)
 {
 	int array_length = 0;
 	struct pvwatchfd_s *info_array = NULL;
+	bool changes_made;
 
 #ifdef __APPLE__
 	struct proc_fdinfo *fd_infos = NULL;
 	int fd_infos_count = 0;
 
 	if (pidfds(state, watch_pid, &fd_infos, &fd_infos_count) != 0) {
-		pv_error(state, "%s: pidfds failed", _("pid"));
+		pv_error("%s: pidfds failed", _("pid"));
 		return -1;
 	}
 #else
@@ -401,18 +455,19 @@ int pv_watchpid_scanfds(pvstate_t state,
 	array_length = *array_length_ptr;
 	info_array = *info_array_ptr;
 
+	changes_made = false;
+
 #ifdef __APPLE__
 	if (fd_infos_count < 1) {
-		pv_error(state, "%s: no fds found", _("pid"));
+		pv_error("%s: no fds found", _("pid"));
 		return -1;
 	}
 	for (int i = 0; i < fd_infos_count; i++) {
 #else
 	while ((d = readdir(dptr)) != NULL) {
 #endif
-		int fd, check_idx, use_idx, rc;
+		int fd, check_idx, found_idx, use_idx, rc;
 		off_t position_now;
-		const char *use_format_string;
 
 		fd = -1;
 #ifdef __APPLE__
@@ -422,24 +477,35 @@ int pv_watchpid_scanfds(pvstate_t state,
 			continue;
 #endif
 
-		/* Skip if the fd is outside the array. */
-		if ((fd < 0) || (fd >= FD_SETSIZE))
+		/* Skip if the fd is negative. */
+		if (fd < 0)
+			continue;
+
+		/* If a watch_fd was specified, skip if this isn't it. */
+		if (watch_fd >= 0 && watch_fd != fd)
 			continue;
 
 		/*
 		 * Skip if this fd is already known to us.
 		 */
-		if (fd_to_idx[fd] != -1) {
-			continue;
+		found_idx = -1;
+		for (check_idx = 0; check_idx < array_length; check_idx++) {
+			if (info_array[check_idx].unused)
+				continue;
+			if (info_array[check_idx].watch_fd != fd)
+				continue;
+			found_idx = check_idx;
+			break;
 		}
+		if (found_idx >= 0)
+			continue;
 
 		/*
-		 * See if there's an empty slot we can re-use. An empty slot
-		 * has a watch_pid of 0.
+		 * See if there's an empty slot we can re-use.
 		 */
 		use_idx = -1;
 		for (check_idx = 0; check_idx < array_length; check_idx++) {
-			if (info_array[check_idx].watch_pid == 0) {
+			if (info_array[check_idx].unused) {
 				use_idx = check_idx;
 				break;
 			}
@@ -458,69 +524,25 @@ int pv_watchpid_scanfds(pvstate_t state,
 
 		debug("%s: %d => index %d", "found new fd", fd, use_idx);
 
+		changes_made = true;
+
 		/*
 		 * Initialise the details of this new entry.
 		 */
 		memset(&(info_array[use_idx]), 0, sizeof(info_array[use_idx]));
 
+		pv_reset_watchfd(&(info_array[use_idx]));
 		info_array[use_idx].watch_pid = watch_pid;
 		info_array[use_idx].watch_fd = fd;
-
-		/* Allocate new display state. */
-		/*@-mustfreeonly@ *//* splint - this is not a leak, this is a new entry. */
-		info_array[use_idx].state = pv_state_alloc(state->status.program_name);
-		/*@+mustfreeonly@ */
-		if (NULL == info_array[use_idx].state)
-			return 2;
-
-		/* Copy the main status.cwd value to the new state. */
-		memcpy(&(info_array[use_idx].state->status.cwd), &(state->status.cwd), sizeof(state->status.cwd));	/* flawfinder: ignore */
-		/* flawfinder: bounded to destination object size. */
+		info_array[use_idx].unused = false;
+		info_array[use_idx].displayable = true;
 
 		/*
-		 * Copy over all the control values, blanking out the
-		 * dynamically allocated strings, and setting the default
-		 * format (fixed buffer) to the required format string
-		 * instead of allocating a new dynamic format string for the
-		 * copy (so there's less dynamic allocation going on).
+		 * Set the average rate window so that a new history buffer
+		 * is allocated for this state.
 		 */
-		memcpy(&(info_array[use_idx].state->control), &(state->control), sizeof(state->control));	/* flawfinder: ignore */
-		/* flawfinder: bounded to destination object size. */
-		/*@-mustfreeonly@ *//* splint - this is not a leak, this is a new entry. */
-		info_array[use_idx].state->control.name = NULL;
-		info_array[use_idx].state->control.format_string = NULL;
-		info_array[use_idx].state->control.output_name = NULL;
-		info_array[use_idx].state->control.default_bar_style = NULL;
-		/*@+mustfreeonly@ */
-		info_array[use_idx].state->control.default_format[0] = '\0';
-		use_format_string =
-		    NULL != state->control.format_string ? state->control.format_string : state->control.default_format;
-		if (NULL != use_format_string)
-			(void) pv_snprintf(info_array[use_idx].state->control.default_format, PV_SIZEOF_DEFAULT_FORMAT,
-					   "%.510s", use_format_string);
-
-		/*
-		 * Duplicate the default bar style string, if there is one.
-		 */
-		if (NULL != state->control.default_bar_style) {
-			pv_state_default_bar_style_set(info_array[use_idx].state, state->control.default_bar_style);
-		}
-
-		/*
-		 * Copy over all the display values, blanking out the
-		 * dynamically allocated parts as above; then set the
-		 * average rate window so that a new history buffer is
-		 * allocated for this state.
-		 */
-		memcpy(&(info_array[use_idx].state->display), &(state->display), sizeof(state->display));	/* flawfinder: ignore */
-		/* flawfinder: bounded to destination object size. */
-		/*@-mustfreeonly@ *//* splint - this is not a leak, this is a new entry. */
-		info_array[use_idx].state->display.display_buffer = NULL;
-		info_array[use_idx].state->display.display_buffer_size = 0;
-		info_array[use_idx].state->calc.history = NULL;
-		info_array[use_idx].state->calc.history_len = 0;
-		/*@+mustfreeonly@ */
-		pv_state_average_rate_window_set(info_array[use_idx].state, state->control.average_rate_window);
+		(void) pv_update_calc_average_rate_window(&(info_array[use_idx].calc),
+							  state->control.average_rate_window);
 
 #ifdef __APPLE__
 		if (fd_infos[i].proc_fdtype != PROX_FDTYPE_VNODE) {
@@ -528,70 +550,33 @@ int pv_watchpid_scanfds(pvstate_t state,
 		}
 #endif
 		/* Retrieve the details of this file descriptor. */
-		rc = pv_watchfd_info(state, &(info_array[use_idx]), true);
+		rc = pv_watchfd_info(state, &(info_array[use_idx]), -1 == watch_fd ? true : false);
 
 		/*
 		 * Lookup failed - mark this slot as being free for re-use.
 		 */
 		if ((rc != 0) && (rc != 4)) {
 			debug("%s %d: %s: %d", "fd", fd, "lookup failed - marking slot for re-use", use_idx);
-			info_array[use_idx].watch_pid = 0;
-			if (NULL != info_array[use_idx].state)
-				pv_state_free(info_array[use_idx].state);
-			/*@-mustfreeonly@ *//* not a leak - we've just free()d it. */
-			info_array[use_idx].state = NULL;
-			/*@+mustfreeonly@ */
+			pv_freecontents_watchfd(&(info_array[use_idx]));
+			info_array[use_idx].unused = true;
+			info_array[use_idx].displayable = false;
 			continue;
 		}
 
-		fd_to_idx[fd] = use_idx;
-
 		/*
-		 * Not displayable - set fd to -1 so the main loop doesn't
-		 * show it.
+		 * Not displayable - mark it as such so the main loop
+		 * doesn't show it.
 		 */
 		if (rc != 0) {
 			debug("%s %d: %s", "fd", fd, "marking as not displayable");
-			info_array[use_idx].watch_fd = -1;
-		}
-
-		/* NULL check on the state - skip if not usable. */
-		if (NULL == info_array[use_idx].state) {
-			debug("%s %d: %s", "fd", fd, "state is NULL - marking as not displayable");
-			info_array[use_idx].watch_fd = -1;
-			continue;
-		}
-
-		/*
-		 * Set the displayed size appropriately; if the size is 0,
-		 * or not known, remove %e and %I from the state's default
-		 * format string so that neither estimated time remaining
-		 * nor estimated time of completion are displayed.
-		 */
-		info_array[use_idx].state->control.size = info_array[use_idx].size;
-		if (info_array[use_idx].state->control.size < 1) {
-			char *fmt;
-			while (NULL != (fmt = strstr(info_array[use_idx].state->control.default_format, "%e"))) {
-				debug("%s", "zero size - removing estimated time remaining");
-				/* strlen-1 here to include trailing \0 */
-				memmove(fmt, fmt + 2, strlen(fmt) - 1);	/* flawfinder: ignore */
-				info_array[use_idx].state->flag.reparse_display = 1;
-			}
-			while (NULL != (fmt = strstr(info_array[use_idx].state->control.default_format, "%I"))) {
-				debug("%s", "zero size - removing estimated completion time");
-				/* strlen-1 here to include trailing \0 */
-				memmove(fmt, fmt + 2, strlen(fmt) - 1);	/* flawfinder: ignore */
-				info_array[use_idx].state->flag.reparse_display = 1;
-			}
-			/* flawfinder: default_format is always \0-terminated. */
+			info_array[use_idx].displayable = false;
 		}
 
 		/* Set the info display_name appropriately. */
 		pv_watchpid_setname(state, &(info_array[use_idx]));
 
-		/* Carry the display_name through to the display state, and reparse. */
-		pv_state_name_set(info_array[use_idx].state, info_array[use_idx].display_name);
-		info_array[use_idx].state->flag.reparse_display = 1;
+		/* Force the display to be re-parsed. */
+		info_array[use_idx].flags.reparse_display = 1;
 
 		pv_elapsedtime_read(&(info_array[use_idx].start_time));
 
@@ -600,11 +585,11 @@ int pv_watchpid_scanfds(pvstate_t state,
 		 * display), if known, so that ETA and so on are calculated
 		 * correctly.
 		 */
-		info_array[use_idx].state->display.initial_offset = 0;
+		info_array[use_idx].display.initial_offset = 0;
 		info_array[use_idx].position = 0;
 		position_now = pv_watchfd_position(&(info_array[use_idx]));
 		if (position_now >= 0) {
-			info_array[use_idx].state->display.initial_offset = position_now;
+			info_array[use_idx].display.initial_offset = position_now;
 			info_array[use_idx].position = position_now;
 		}
 	}
@@ -616,6 +601,14 @@ int pv_watchpid_scanfds(pvstate_t state,
 	(void) closedir(dptr);
 #endif
 
+	/*
+	 * If any changes were made (i.e. new file descriptors were found),
+	 * sort the array so that file descriptors are always displayed in
+	 * ascending numerical order.
+	 */
+	if (changes_made && NULL != info_array && array_length > 1)
+		qsort(info_array, (size_t) array_length, sizeof(info_array[0]), pv_compare_watchfd);
+
 	return 0;
 }
 
@@ -624,6 +617,9 @@ int pv_watchpid_scanfds(pvstate_t state,
  * Set the display name for the given watched file descriptor, truncating at
  * the relevant places according to the current screen width.
  *
+ * If more than one distinct PID is being watched, include the PID in the
+ * name as well as the file descriptor number.
+ *
  * If the file descriptor is pointing to a file under the current working
  * directory, show its relative path, not the full path.
  */
@@ -646,20 +642,36 @@ void pv_watchpid_setname(pvstate_t state, pvwatchfd_t info)
 	}
 
 	max_display_length = (int) (state->control.width / 2) - 6;
+	if (state->watchfd.multiple_pids)
+		max_display_length -= 9;
 	if (max_display_length >= (int) path_length) {
-		(void) pv_snprintf(info->display_name,
-				   PV_SIZEOF_DISPLAY_NAME, "%4d:%.498s", info->watch_fd, file_fdpath);
+		if (state->watchfd.multiple_pids) {
+			(void) pv_snprintf(info->display_name,
+					   PV_SIZEOF_DISPLAY_NAME, "%8d:%4d:%.498s", (int) (info->watch_pid),
+					   info->watch_fd, file_fdpath);
+		} else {
+			(void) pv_snprintf(info->display_name,
+					   PV_SIZEOF_DISPLAY_NAME, "%4d:%.498s", info->watch_fd, file_fdpath);
+		}
 	} else {
 		int prefix_length, suffix_length;
 
 		prefix_length = max_display_length / 4;
 		suffix_length = max_display_length - prefix_length - 3;
 
-		(void) pv_snprintf(info->display_name,
-				   PV_SIZEOF_DISPLAY_NAME,
-				   "%4d:%.*s...%.*s",
-				   info->watch_fd, prefix_length,
-				   file_fdpath, suffix_length, file_fdpath + path_length - suffix_length);
+		if (state->watchfd.multiple_pids) {
+			(void) pv_snprintf(info->display_name,
+					   PV_SIZEOF_DISPLAY_NAME,
+					   "%8d:%4d:%.*s...%.*s",
+					   (int) (info->watch_pid), info->watch_fd, prefix_length,
+					   file_fdpath, suffix_length, file_fdpath + path_length - suffix_length);
+		} else {
+			(void) pv_snprintf(info->display_name,
+					   PV_SIZEOF_DISPLAY_NAME,
+					   "%4d:%.*s...%.*s",
+					   info->watch_fd, prefix_length,
+					   file_fdpath, suffix_length, file_fdpath + path_length - suffix_length);
+		}
 	}
 
 	debug("%s: %d: [%s]", "set name for fd", info->watch_fd, info->display_name);
